"""
    hiku.readers.protobuf
    ~~~~~~~~~~~~~~~~~~~~~

    Support for queries encoded using Protocol Buffers

"""
from ..query import Node, Link, Field
from ..protobuf import query_pb2


_OPTION_VALUE_GETTERS = {
    'string': lambda op: op.string,
    'integer': lambda op: op.integer,
    'repeated_string': lambda op: list(op.repeated_string.items),
    'repeated_integer': lambda op: list(op.repeated_integer.items),
}


def _transform_options(pb_obj):
    options = {}
    for key, pb_option in pb_obj.options.items():
        option_type = pb_option.WhichOneof('value')
        if option_type is None:
            raise TypeError('Option value is not set: {!r}'.format(pb_option))
        options[key] = _OPTION_VALUE_GETTERS[option_type](pb_option)
    return options or None


def transform(pb_node):
    fields = []
    for i in pb_node.items:
        item_type = i.WhichOneof('value')
        if item_type == 'field':
            if not i.field.name:
                raise TypeError('Field name is empty: {!r}'.format(i))
            fields.append(Field(i.field.name, _transform_options(i.field)))
        elif item_type == 'link':
            if not i.link.name:
                raise TypeError('Link name is empty: {!r}'.format(i))
            fields.append(Link(i.link.name, transform(i.link.node),
                               _transform_options(i.link)))
        else:
            raise TypeError('Node item is empty: {!r}'.format(i))
    return Node(fields)


def read(data):
    """Reads a query, encoded into binary Protocol Buffers format, using
    message types from the ``hiku.protobuf.query`` package.

    Proto-file location: ``hiku/protobuf/query.proto``

    Generated message types: ``hiku.protobuf.query_pb2``

    Example:

    .. code-block:: python

        from hiku.builder import Q, build
        from hiku.export.protobuf import export

        bin_query = export(build([
            Q.characters(limit=100)[
                Q.name,
            ],
        ])).SerializeToString()

        assert bin_query == (
            b'\\n(\\x12&\\n\\ncharacters\\x12\\n\\n\\x08\\n\\x06\\n\\x04'
            b'name\\x1a\\x0c\\n\\x05limit\\x12\\x03\\x10\\xc8\\x01'
        )

        query = read(bin_query)  # reading binary message

        result = engine.execute(graph, query)

    :param bytes data: binary message representation
    :return: :py:class:`hiku.query.Node`, ready to execute query object
    """
    pb_value = query_pb2.Node.FromString(data)
    return transform(pb_value)



"""
    hiku.result
    ~~~~~~~~~~~

    In all examples query results are showed in **denormalized** form, suitable
    for reading (by humans) and for serializing into simple formats, into `JSON`
    for example. But this is not how `Hiku` stores result internally.

    Internally `Hiku` stores result in a fully **normalized** form. So result in
    `Hiku` is also a graph structure with references between objects. This
    approach has lots of advantages:

      - normalization helps to heavily reduce size of serialized result when we
        need to transfer it (this avoids data duplication)
      - it reduces internal memory usage and simplifies work with data
        internally
      - gives ability to cache, precisely and effortlessly update local state
        on the client

"""
from .types import RecordMeta, OptionalMeta, SequenceMeta
from .query import Node, Field, Link, merge
from .graph import Link as GraphLink, Field as GraphField, Many, Maybe


class Ref(object):

    def __init__(self, index, node, ident):
        self.index = index
        self.node = node
        self.ident = ident

    def __getitem__(self, key):
        return self.index[self.node][self.ident][key]

    def __repr__(self):
        return '<{}:{}>'.format(self.node, self.ident)

    def __eq__(self, other):
        return self.index[self.node][self.ident] == other


class Result(object):
    """Internal result representation

    It gives access to the result of the :py:class:`~hiku.graph.Root` node --
    which is a starting point of the query execution, and from it to all the
    other node objects, which are stored internally in the index.

    Behaves like a mapping.
    """
    def __init__(self):
        self.root = {}
        self.index = {}

    def __getitem__(self, key):
        return self.root[key]

    def ref(self, node, ident):
        return Ref(self.index, node, ident)


def _filter_fields(result, node):
    return {f.name: result[f.name] for f in node.fields}


def _denormalize(graph, graph_obj, result, query_obj):
    if isinstance(query_obj, Node):
        return {f.name: _denormalize(graph, graph_obj.fields_map[f.name],
                                     result[f.name], f)
                for f in query_obj.fields}

    elif isinstance(query_obj, Field):
        return result

    elif isinstance(query_obj, Link):
        if isinstance(graph_obj, GraphField):
            type_ = graph_obj.type
            if isinstance(type_, SequenceMeta):
                return [_filter_fields(item, query_obj.node) for item in result]
            elif isinstance(type_, OptionalMeta):
                return (_filter_fields(result, query_obj.node)
                        if result is not None else None)
            else:
                assert isinstance(type_, RecordMeta), repr(type_)
                return _filter_fields(result, query_obj.node)

        elif isinstance(graph_obj, GraphLink):
            graph_node = graph.nodes_map[graph_obj.node]
            if graph_obj.type_enum is Many:
                return [_denormalize(graph, graph_node, v, query_obj.node)
                        for v in result]
            elif graph_obj.type_enum is Maybe and result is None:
                return None
            else:
                return _denormalize(graph, graph_node, result, query_obj.node)

        else:
            return _denormalize(graph, graph_obj, result, query_obj.node)


def denormalize(graph, result, query):
    """Transforms normalized result (graph) into simple hierarchical structure

    This hierarchical structure will follow query structure.

    Example::

        query = hiku.readers.simple.read('[:foo]')
        norm_result = hiku_engine.execute(graph, query)
        result = hiku.result.denormalize(graph, norm_result, query)
        assert result == {'foo': 'value'}

    :param graph: :py:class:`~hiku.graph.Graph` definition
    :param result: result of the query
    :param query: executed query, instance of the :py:class:`~hiku.query.Node`
    """
    return _denormalize(graph, graph.root, result, merge([query]))



from ..query import Node, Link, Field, merge
from ..types import GenericMeta, RecordMeta, SequenceMeta, MappingMeta
from ..types import OptionalMeta, TypeRefMeta

from .nodes import NodeVisitor


# TODO: revisit this
_CONTAINER_TYPES = (
    OptionalMeta,
    SequenceMeta,
    MappingMeta,
    RecordMeta,
)


class Ref(object):

    def __init__(self, backref, to):
        self.backref = backref
        self.to = to


class NamedRef(Ref):

    def __init__(self, backref, name, to):
        super(NamedRef, self).__init__(backref, to)
        self.name = name


def get_type(types, type_):
    return (types[type_.__type_name__]
            if isinstance(type_, TypeRefMeta)
            else type_)


def ref_to_req(types, ref, add_req=None):
    if ref is None:
        assert add_req is not None
        return add_req

    ref_type = get_type(types, ref.to)
    if isinstance(ref_type, OptionalMeta):
        ref_type = get_type(types, ref_type.__type__)

    if isinstance(ref_type, RecordMeta):
        if isinstance(ref, NamedRef):
            node = Node([]) if add_req is None else add_req
            return ref_to_req(types, ref.backref,
                              Node([Link(ref.name, node)]))
        else:
            return ref_to_req(types, ref.backref, add_req)

    elif isinstance(ref_type, SequenceMeta):
        item_type = get_type(types, ref_type.__item_type__)
        if isinstance(item_type, RecordMeta):
            assert isinstance(ref, NamedRef), type(ref)
            node = Node([]) if add_req is None else add_req
            return ref_to_req(types, ref.backref,
                              Node([Link(ref.name, node)]))
        else:
            assert not isinstance(item_type, _CONTAINER_TYPES), ref_type
            assert isinstance(ref, NamedRef), type(ref)
            assert add_req is None, repr(add_req)
            return ref_to_req(types, ref.backref,
                              Node([Field(ref.name)]))

    elif isinstance(ref_type, GenericMeta):
        assert not isinstance(ref_type, _CONTAINER_TYPES), ref_type
        assert isinstance(ref, NamedRef), type(ref)
        assert add_req is None, repr(add_req)
        return ref_to_req(types, ref.backref,
                          Node([Field(ref.name)]))

    else:
        raise TypeError('Reference to the invalid type: {!r}'
                        .format(ref_type))


def type_to_query(type_):
    fields = []
    for f_name, f_type in type_.__field_types__.items():
        if isinstance(f_type, OptionalMeta):
            f_type = f_type.__type__
        if isinstance(f_type, RecordMeta):
            fields.append(Link(f_name, type_to_query(f_type)))
        elif isinstance(f_type, SequenceMeta):
            if isinstance(f_type.__item_type__, RecordMeta):
                fields.append(Link(f_name, type_to_query(f_type.__item_type__)))
            else:
                raise NotImplementedError(type(f_type.__item_type__))
        else:
            fields.append(Field(f_name))
    return Node(fields)


class RequirementsExtractor(NodeVisitor):

    def __init__(self, types):
        self._types = types
        self._reqs = []

    @classmethod
    def extract(cls, types, expr):
        extractor = cls(types)
        extractor.visit(expr)
        return merge(extractor._reqs)

    def visit(self, node):
        ref = getattr(node, '__ref__', None)
        if ref is not None:
            req = ref_to_req(self._types, ref)
            if req is not None:
                self._reqs.append(req)
        super(RequirementsExtractor, self).visit(node)

    def visit_tuple(self, node):
        sym, args = node.values[0], node.values[1:]
        sym_ref = getattr(sym, '__ref__', None)
        if sym_ref is not None:
            for arg, arg_type in zip(args, sym_ref.to.__arg_types__):
                if isinstance(arg_type, OptionalMeta):
                    arg_type = arg_type.__type__
                if isinstance(arg_type, RecordMeta):
                    self._reqs.append(ref_to_req(self._types, arg.__ref__,
                                                 type_to_query(arg_type)))
                else:
                    self.visit(arg)
        else:
            for arg in args:
                self.visit(arg)



from itertools import chain
from collections import defaultdict


class Workflow(object):

    def result(self):
        raise NotImplementedError(type(self))


class TaskSet(object):

    def __init__(self, queue):
        self._queue = queue

    def submit(self, fn, *args, **kwargs):
        return self._queue.submit(self, fn, *args, **kwargs)


class Queue(object):

    def __init__(self, executor):
        self._executor = executor
        self._futures = {}
        self._forks = {}
        self._callbacks = defaultdict(list)

    @property
    def __futures__(self):
        return list(chain.from_iterable(self._futures.values()))

    def progress(self, done):
        for future_set in self._futures.values():
            future_set.difference_update(done)

        for fut in done:
            for callback in self._callbacks.pop(fut, []):
                callback()

        while True:
            completed_task_sets = [ts for ts in self._futures.keys()
                                   if not self._futures[ts] and
                                   not self._forks[ts]]
            if not completed_task_sets:
                break

            for task_set in completed_task_sets:
                for callback in self._callbacks.pop(task_set, []):
                    callback()
                self._futures.pop(task_set)
                self._forks.pop(task_set)
                for fork_set in self._forks.values():
                    fork_set.discard(task_set)

    def submit(self, task_set, fn, *args, **kwargs):
        fut = self._executor.submit(fn, *args, **kwargs)
        self._futures[task_set].add(fut)
        return fut

    def fork(self, from_):
        task_set = TaskSet(self)
        self._futures[task_set] = set()
        self._forks[task_set] = set()
        if from_ is not None:
            self._forks[from_].add(task_set)
        return task_set

    def add_callback(self, obj, callback):
        self._callbacks[obj].append(callback)



"""
    hiku.expr.core
    ~~~~~~~~~~~~~~

    Expression building blocks

"""
from functools import wraps
from itertools import chain
from collections import namedtuple

from ..edn import loads
from ..query import Node, Link, Field
from ..types import Record, Callable, Any
from ..compat import text_type, string_types
from ..readers.simple import transform

from .nodes import Symbol, Tuple, List, Keyword, Dict


_Func = namedtuple('__func__', 'expr, args')


class _DotHandler(object):

    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return repr(self.obj)

    def __getattr__(self, name):
        return _DotHandler(Tuple([Symbol('get'), self.obj, Symbol(name)]))


class _S(object):

    def __getattr__(self, name):
        return _DotHandler(Symbol(name))


#: Helper object to represent symbols in expressions. ``S.foo.bar`` in
#: expressions is equivalent to ``foo.bar`` in the regular Python.
S = _S()


def _to_expr(obj, fn_reg):
    if isinstance(obj, _DotHandler):
        return obj.obj
    elif isinstance(obj, _Func):
        fn_reg.add(obj.expr)
        return Tuple([Symbol(obj.expr.__def_name__)] +
                     [_to_expr(arg, fn_reg) for arg in obj.args])
    elif isinstance(obj, list):
        return List(_to_expr(v, fn_reg) for v in obj)
    elif isinstance(obj, dict):
        values = chain.from_iterable((Keyword(k), _to_expr(v, fn_reg))
                                     for k, v in obj.items())
        return Dict(values)
    else:
        return obj


def to_expr(obj):
    functions = set([])
    node = _to_expr(obj, functions)
    return node, tuple(functions)


def _query_to_types(obj):
    if isinstance(obj, Node):
        return Record[[(f.name, _query_to_types(f)) for f in obj.fields]]
    elif isinstance(obj, Link):
        return _query_to_types(obj.node)
    elif isinstance(obj, Field):
        return Any
    else:
        raise TypeError(type(obj))


def define(*types, **kwargs):
    """Annotates function arguments with types.

    These annotations are used to type-check expressions and to analyze,
    which data is used from provided arguments.

    Example:

    .. code-block:: python

        @define(Record[{'id': Integer, 'name': String}])
        def image_url(image):
            return 'http://example.com/{id}-{name}'.format(id=image['id'],
                                                           name=image['name'])

    Here ``image_url`` function accepts an object as argument, and is using two
    of it's fields: ``id`` field of type ``Integer`` and ``name`` field of type
    ``String``. Hiku will check that this function will be used only with
    objects having at least such two fields.

    This annotation also gives ability for Hiku to build a query for low-level
    graph.
    """
    def decorator(fn):
        _name = kwargs.pop('_name', None)
        assert not kwargs, repr(kwargs)

        name = _name or '{}/{}_{}'.format(fn.__module__, fn.__name__, id(fn))

        @wraps(fn)
        def expr(*args):
            return _Func(expr, args)

        expr.__def_name__ = name
        expr.__def_body__ = fn

        if len(types) == 1 and isinstance(types[0], string_types):
            reqs_list = loads(text_type(types[0]))
            expr.__def_type__ = Callable[[(_query_to_types(transform(r))
                                           if r is not None else Any)
                                          for r in reqs_list]]
        else:
            expr.__def_type__ = Callable[types]

        return expr
    return decorator


@define(Any, Any, Any, _name='each')
def each(var, col, expr):
    """Returns a list of the results of the expression evaluation for every
    item of the sequence provided.

    Example:

    .. code-block:: python

        each(S.x, S.collection, S.x.name)

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        [x.name for x in collection]

    """


@define(Any, Any, Any, _name='if')
def if_(test, then, else_):
    """Checks condition and continues to evaluate one of the two expressions
    provided.

    Example:

    .. code-block:: python

        if_(S.value, 'truish', 'falsish')

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        if value:
            return 'truish'
        else:
            return 'falsish'

    """


@define(Any, Any, Any, _name='if_some')
def if_some(bind, then, else_):
    """Used to unpack values with ``Optional`` types and using them safely in
    expressions.

    Example:

    .. code-block:: python

        if_some([S.img, S.this.image],
                image_url(S.img),
                'http://example.com/no-photo.jpg')

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        if this.image is not None:
            img = this.image
            return image_url(img)
        else:
            return 'http://example.com/no-photo.jpg'

    If ``S.this.image`` has a type ``Optional[TypeRef['Image']]``, ``S.img``
    variable will have a type ``TypeRef['Image']`` and it will be available only
    in "then"-expression, which will be evaluated only if ``S.this.image``
    wouldn't be ``None``. Otherwise "else"-expression will be
    evaluated.
    """



from ..types import SequenceMeta, OptionalMeta
from ..query import Field, Link, merge
from ..graph import Node as GraphNode, Link as GraphLink, Field as GraphField
from ..graph import Many, Maybe


def _set_fields(pb_obj, result, node):
    for f in node.fields:
        setattr(pb_obj, f.name, result[f.name])


def _populate_node(pb_node, graph, graph_node, result, query_node):
    for f in query_node.fields:
        if isinstance(f, Field):
            setattr(pb_node, f.name, result[f.name])
        elif isinstance(f, Link):
            _transform_link(getattr(pb_node, f.name), graph,
                            graph_node.fields_map[f.name], result[f.name], f)
        else:
            raise TypeError(type(f))


def _transform_link(pb_obj, graph, graph_obj, result, query_obj):
    # complex field type
    if isinstance(graph_obj, GraphField):
        type_ = graph_obj.type
        if isinstance(type_, SequenceMeta):
            for item in result:
                pb_item = pb_obj.add()
                _set_fields(pb_item, item, query_obj.node)
        elif isinstance(type_, OptionalMeta) and result is None:
            pass
        else:
            _set_fields(pb_obj, result, query_obj.node)

    # regular link to node
    elif isinstance(graph_obj, GraphLink):
        graph_node = graph.nodes_map[graph_obj.node]
        if graph_obj.type_enum is Many:
            for v in result:
                pb_item = pb_obj.add()
                _populate_node(pb_item, graph, graph_node, v, query_obj.node)
        elif graph_obj.type_enum is Maybe and result is None:
            pass
        else:
            _populate_node(pb_obj, graph, graph_node, result, query_obj.node)

    # node in root node
    else:
        assert isinstance(graph_obj, GraphNode)
        _populate_node(pb_obj, graph, graph_obj, result, query_obj.node)


def populate(pb_root, graph, result, query):
    _populate_node(pb_root, graph, graph.root, result, merge([query]))



from abc import abstractmethod, ABCMeta
from collections import OrderedDict

from .compat import with_metaclass


class GenericMeta(type):

    def __repr__(cls):
        return cls.__name__

    def accept(cls, visitor):
        raise NotImplementedError(type(cls))


class AnyMeta(GenericMeta):

    def accept(cls, visitor):
        return visitor.visit_any(cls)


class Any(with_metaclass(AnyMeta, object)):
    pass


class BooleanMeta(GenericMeta):

    def accept(cls, visitor):
        return visitor.visit_boolean(cls)


class Boolean(with_metaclass(BooleanMeta, object)):
    pass


class StringMeta(GenericMeta):

    def accept(cls, visitor):
        return visitor.visit_string(cls)


class String(with_metaclass(StringMeta, object)):
    pass


class IntegerMeta(GenericMeta):

    def accept(cls, visitor):
        return visitor.visit_integer(cls)


class Integer(with_metaclass(IntegerMeta, object)):
    pass


class FloatMeta(GenericMeta):

    def accept(cls, visitor):
        return visitor.visit_float(cls)


class Float(with_metaclass(FloatMeta, object)):
    pass


class TypingMeta(GenericMeta):
    __final__ = False

    def __cls_init__(cls, *args):
        raise NotImplementedError(type(cls))

    def __cls_repr__(cls):
        raise NotImplementedError(type(cls))

    def __getitem__(cls, parameters):
        if cls.__final__:
            raise TypeError('Cannot substitute parameters in {!r}'.format(cls))
        type_ = cls.__class__(cls.__name__, cls.__bases__, dict(cls.__dict__))
        type_.__cls_init__(parameters)
        type_.__final__ = True
        return type_

    def __repr__(self):
        if self.__final__:
            return self.__cls_repr__()
        else:
            return super(TypingMeta, self).__repr__()


class OptionalMeta(TypingMeta):

    def __cls_init__(cls, type_):
        cls.__type__ = type_

    def __cls_repr__(self):
        return '{}[{!r}]'.format(self.__name__, self.__type__)

    def accept(cls, visitor):
        return visitor.visit_optional(cls)


class Optional(with_metaclass(OptionalMeta, object)):
    pass


class SequenceMeta(TypingMeta):

    def __cls_init__(cls, item_type):
        cls.__item_type__ = item_type

    def __cls_repr__(self):
        return '{}[{!r}]'.format(self.__name__, self.__item_type__)

    def accept(cls, visitor):
        return visitor.visit_sequence(cls)


class Sequence(with_metaclass(SequenceMeta, object)):
    pass


class MappingMeta(TypingMeta):

    def __cls_init__(cls, params):
        cls.__key_type__, cls.__value_type__ = params

    def __cls_repr__(self):
        return '{}[{!r}, {!r}]'.format(self.__name__, self.__key_type__,
                                       self.__value_type__)

    def accept(cls, visitor):
        return visitor.visit_mapping(cls)


class Mapping(with_metaclass(MappingMeta, object)):
    pass


class RecordMeta(TypingMeta):

    def __cls_init__(cls, field_types):
        cls.__field_types__ = OrderedDict(field_types)

    def __cls_repr__(self):
        return '{}[{!r}]'.format(self.__name__, dict(self.__field_types__))

    def accept(cls, visitor):
        return visitor.visit_record(cls)


class Record(with_metaclass(RecordMeta, object)):
    pass


class CallableMeta(TypingMeta):

    def __cls_init__(cls, arg_types):
        cls.__arg_types__ = arg_types

    def __cls_repr__(self):
        return '{}[{}]'.format(self.__name__,
                               ', '.join(map(repr, self.__arg_types__)))

    def accept(cls, visitor):
        return visitor.visit_callable(cls)


class Callable(with_metaclass(CallableMeta, object)):
    pass


class TypeRefMeta(TypingMeta):

    def __cls_init__(cls, name):
        cls.__type_name__ = name

    def __cls_repr__(self):
        return '{}[{!r}]'.format(self.__name__, self.__type_name__)

    def accept(cls, visitor):
        return visitor.visit_typeref(cls)


class TypeRef(with_metaclass(TypeRefMeta, object)):
    pass


class AbstractTypeVisitor(with_metaclass(ABCMeta, object)):

    @abstractmethod
    def visit(self, obj):
        pass

    @abstractmethod
    def visit_any(self, obj):
        pass

    @abstractmethod
    def visit_boolean(self, obj):
        pass

    @abstractmethod
    def visit_string(self, obj):
        pass

    @abstractmethod
    def visit_integer(self, obj):
        pass

    @abstractmethod
    def visit_float(self, obj):
        pass

    @abstractmethod
    def visit_typeref(self, obj):
        pass

    @abstractmethod
    def visit_optional(self, obj):
        pass

    @abstractmethod
    def visit_sequence(self, obj):
        pass

    @abstractmethod
    def visit_mapping(self, obj):
        pass

    @abstractmethod
    def visit_record(self, obj):
        pass

    @abstractmethod
    def visit_callable(self, obj):
        pass


class TypeVisitor(AbstractTypeVisitor):

    def visit(self, obj):
        return obj.accept(self)

    def visit_any(self, obj):
        pass

    def visit_boolean(self, obj):
        pass

    def visit_string(self, obj):
        pass

    def visit_integer(self, obj):
        pass

    def visit_float(self, obj):
        pass

    def visit_typeref(self, obj):
        pass

    def visit_optional(self, obj):
        self.visit(obj.__type__)

    def visit_sequence(self, obj):
        self.visit(obj.__item_type__)

    def visit_mapping(self, obj):
        self.visit(obj.__key_type__)
        self.visit(obj.__value_type__)

    def visit_record(self, obj):
        for value_type in obj.__field_types__.values():
            self.visit(value_type)

    def visit_callable(self, obj):
        for arg_type in obj.__arg_types__:
            self.visit(arg_type)



class Node(object):

    def accept(self, visitor):
        raise NotImplementedError(type(self))


class Symbol(Node):

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name

    def accept(self, visitor):
        return visitor.visit_symbol(self)


class Keyword(Node):

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return ':{}'.format(self.name)

    def accept(self, visitor):
        return visitor.visit_keyword(self)


class Tuple(Node):

    def __init__(self, values):
        self.values = tuple(values)

    def __repr__(self):
        return '({})'.format(' '.join(map(repr, self.values)))

    def accept(self, visitor):
        return visitor.visit_tuple(self)


class List(Node):

    def __init__(self, values):
        self.values = tuple(values)

    def __repr__(self):
        return '[{}]'.format(' '.join(map(repr, self.values)))

    def accept(self, visitor):
        return visitor.visit_list(self)


class Dict(Node):

    def __init__(self, values):
        self.values = tuple(values)

    def __repr__(self):
        return '{{{}}}'.format(' '.join(map(repr, self.values)))

    def accept(self, visitor):
        return visitor.visit_dict(self)


class NodeVisitor(object):

    def visit(self, node):
        if hasattr(node, 'accept'):
            node.accept(self)
        else:
            self.generic_visit(node)

    def generic_visit(self, node):
        pass

    def visit_symbol(self, node):
        pass

    def visit_keyword(self, node):
        pass

    def visit_tuple(self, node):
        for value in node.values:
            self.visit(value)

    def visit_list(self, node):
        for value in node.values:
            self.visit(value)

    def visit_dict(self, node):
        for value in node.values:
            self.visit(value)


class NodeTransformer(object):

    def visit(self, node):
        if hasattr(node, 'accept'):
            return node.accept(self)
        else:
            return self.generic_visit(node)

    def generic_visit(self, node):
        return node

    def visit_symbol(self, node):
        return Symbol(node.name)

    def visit_keyword(self, node):
        return Keyword(node.name)

    def visit_tuple(self, node):
        return Tuple(self.visit(value)
                     for value in node.values)

    def visit_list(self, node):
        return List(self.visit(value)
                    for value in node.values)

    def visit_dict(self, node):
        return Dict(self.visit(value)
                    for value in node.values)



from contextlib import contextmanager
from collections import deque, OrderedDict

from .. import graph
from ..types import Sequence, SequenceMeta, Record, RecordMeta
from ..types import MappingMeta, OptionalMeta, Any, AnyMeta, TypeRefMeta

from .refs import NamedRef, Ref
from .nodes import NodeTransformer, Symbol, Keyword, Tuple, List


class GraphTypes(graph.GraphVisitor):

    def visit_graph(self, obj):
        types = OrderedDict((node.name, self.visit(node))
                            for node in obj.iter_nodes())
        types['__root__'] = self.visit(graph.Root(list(obj.iter_root())))
        return types

    def visit_node(self, obj):
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_root(self, obj):
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_link(self, obj):
        return obj.type

    def visit_field(self, obj):
        return obj.type or Any


def graph_types(graph_):
    return GraphTypes().visit(graph_)


def fn_types(functions):
    return {fn.__def_name__: fn.__def_type__ for fn in functions}


class Environ(object):

    def __init__(self, values):
        env_vars = {}
        for name, value in values.items():
            env_vars[name] = NamedRef(None, name, value)
        self.vars = deque([env_vars])

    @contextmanager
    def push(self, mapping):
        self.vars.append(mapping)
        try:
            yield
        finally:
            self.vars.pop()

    def __getitem__(self, key):
        for d in reversed(self.vars):
            try:
                return d[key]
            except KeyError:
                continue
        else:
            raise KeyError(repr(key))

    def __contains__(self, key):
        return any(key in d for d in self.vars)


def get_type(types, obj):
    if isinstance(obj, TypeRefMeta):
        return types[obj.__type_name__]
    else:
        return obj


def node_type(types, node):
    return get_type(types, node.__ref__.to)


def check_type(types, t1, t2):
    t1 = get_type(types, t1)
    t2 = get_type(types, t2)
    if isinstance(t2, AnyMeta):
        pass
    else:
        if isinstance(t1, type(t2)):
            if isinstance(t2, OptionalMeta):
                check_type(types, t1.__type__, t2.__type__)
            elif isinstance(t2, SequenceMeta):
                check_type(types, t1.__item_type__, t2.__item_type__)
            elif isinstance(t2, MappingMeta):
                check_type(types, t1.__key_type__, t2.__key_type__)
                check_type(types, t1.__value_type__, t2.__value_type__)
            elif isinstance(t2, RecordMeta):
                for key, v2 in t2.__field_types__.items():
                    v1 = t1.__field_types__.get(key)
                    if v1 is None:
                        raise TypeError('Missing field {}'.format(key))
                    v1 = get_type(types, v1)
                    check_type(types, v1, v2)
        elif isinstance(t2, OptionalMeta):
            check_type(types, t1, t2.__type__)
        else:
            raise TypeError('Types mismatch, {!r} != {!r}'.format(t1, t2))


class Checker(NodeTransformer):

    def __init__(self, types, env):
        self.types = types
        self.env = Environ(env)

    def visit_get_expr(self, node):
        sym, obj, name = node.values
        assert isinstance(name, Symbol), type(name)
        obj = self.visit(obj)
        assert hasattr(obj, '__ref__'), 'Object does not have a reference'

        ref_to = node_type(self.types, obj)
        check_type(self.types, ref_to, Record[{name.name: Any}])

        res = ref_to.__field_types__.get(name.name)
        assert res is not None, 'Undefined field name: {}'.format(name.name)
        tup = Tuple([sym, obj, name])
        tup.__ref__ = NamedRef(obj.__ref__, name.name, res)
        return tup

    def visit_each_expr(self, node):
        var, col, expr = node.values[1:]
        assert isinstance(var, Symbol), repr(var)
        col = self.visit(col)
        assert hasattr(col, '__ref__'), 'Object does not have a reference'
        col_type = node_type(self.types, col)
        check_type(self.types, col_type, Sequence[Record[{}]])
        var = Symbol(var.name)
        var.__ref__ = Ref(col.__ref__, col_type.__item_type__)
        with self.env.push({var.name: var.__ref__}):
            expr = self.visit(expr)
        return Tuple([Symbol('each'), var, col, expr])

    def visit_if_some_expr(self, node):
        bind, then, else_ = node.values[1:]
        assert isinstance(bind, List) and len(bind.values) == 2, repr(bind)
        bind_sym, bind_expr = bind.values
        assert isinstance(bind_sym, Symbol), repr(bind_sym)
        bind_expr = self.visit(bind_expr)
        assert hasattr(bind_expr, '__ref__'), 'Object does not have a reference'
        bind_type = node_type(self.types, bind_expr)
        if isinstance(bind_type, OptionalMeta):
            bind_expr_ref = Ref(bind_expr.__ref__, bind_type.__type__)
        else:
            # TODO: warn about unnecessary check
            bind_expr_ref = bind_expr.__ref__
        bind_sym = Symbol(bind_sym.name)
        bind_sym.__ref__ = bind_expr_ref
        with self.env.push({bind_sym.name: bind_sym.__ref__}):
            then = self.visit(then)
        else_ = self.visit(else_)
        return Tuple([Symbol('if_some'), List([bind_sym, bind_expr]),
                      then, else_])

    def visit_tuple_generic(self, node):
        sym = self.visit(node.values[0])
        assert isinstance(sym, Symbol), type(sym)
        args = [self.visit(val) for val in node.values[1:]]
        fn_type = node_type(self.types, sym)
        assert len(fn_type.__arg_types__) == len(args), 'Wrong arguments count'
        for arg, arg_type in zip(args, fn_type.__arg_types__):
            ref = getattr(arg, '__ref__', None)
            if ref is not None:
                check_type(self.types, node_type(self.types, arg), arg_type)
            else:
                check_type(self.types, Any, arg_type)
        return Tuple([sym] + args)

    def visit_tuple(self, node):
        sym = node.values[0]
        if sym.name == 'get':
            return self.visit_get_expr(node)
        elif sym.name == 'each':
            return self.visit_each_expr(node)
        elif sym.name == 'if_some':
            return self.visit_if_some_expr(node)
        else:
            return self.visit_tuple_generic(node)

    def visit_symbol(self, node):
        sym = Symbol(node.name)
        assert node.name in self.env, 'Unknown symbol {}'.format(node.name)
        sym.__ref__ = self.env[node.name]
        return sym

    def visit_dict(self, node):
        assert not len(node.values) % 2, 'Probably missing keyword value'
        keys = node.values[::2]
        assert all(isinstance(k, Keyword) for k in keys), 'Wrong arguments'
        return super(Checker, self).visit_dict(node)


def check(expr, types, env):
    return Checker(types, env).visit(expr)



from contextlib import contextmanager
from collections import Counter

from ..types import CallableMeta
from ..compat import ast as py, text_type, PY36

from .nodes import Symbol, Keyword


class Environ(object):

    def __init__(self):
        self.vars = Counter()

    def __getitem__(self, key):
        i = self.vars[key]
        return '{}_{}'.format(key, i) if i > 1 else key

    def __contains__(self, key):
        return key in self.vars

    @contextmanager
    def push(self, names):
        for name in names:
            self.vars[name] += 1
        try:
            yield
        finally:
            for name in names:
                self.vars[name] -= 1


class ExpressionCompiler(object):
    ctx_var = 'ctx'
    env_var = 'env'

    def __init__(self):
        self.env = Environ()

    @classmethod
    def compile_expr(cls, node):
        return cls().visit(node)

    @classmethod
    def compile_lambda_expr(cls, node, args=None):
        args = args or []
        compiler = cls()
        with compiler.env.push(['this'] + args):
            body = compiler.visit(node)
        py_args = [py.arg(cls.env_var), py.arg('this'), py.arg(cls.ctx_var)]
        py_args += [py.arg(name) for name in args]
        expr = py.Lambda(py.arguments(py_args, None, None, []), body)
        py.fix_missing_locations(expr)
        return py.Expression(expr)

    def _var_load(self, name):
        return py.Name(name, py.Load())

    def _ctx_load(self, name):
        return py.Subscript(py.Name(self.ctx_var, py.Load()),
                            py.Index(py.Str(name)), py.Load())

    def _env_load(self, name):
        return py.Subscript(py.Name(self.env_var, py.Load()),
                            py.Index(py.Str(name)), py.Load())

    def visit(self, node):
        if hasattr(node, 'accept'):
            return node.accept(self)
        else:
            return self.generic_visit(node)

    def generic_visit(self, node):
        if isinstance(node, int):
            return py.Num(node)
        elif isinstance(node, text_type):
            return py.Str(node)
        else:
            raise NotImplementedError(type(node))

    def visit_get_expr(self, node):
        _, obj, name = node.values
        assert isinstance(name, Symbol)
        obj_expr = self.visit(obj)
        return py.Subscript(obj_expr, py.Index(py.Str(name.name)), py.Load())

    def visit_if_expr(self, node):
        args = [self.visit(val) for val in node.values[1:]]
        test_expr, then_expr, else_expr = args
        return py.IfExp(test_expr, then_expr, else_expr)

    def visit_if_some_expr(self, node):
        _, bind, then_, else_ = node.values
        bind_sym, bind_expr = bind.values
        with self.env.push([bind_sym.name]):
            none = py.Name('None', py.Load())
            load_bind_sym = py.Name(self.env[bind_sym.name], py.Load())
            test = py.Compare(load_bind_sym, [py.IsNot()], [none])
            store_bind_sym = py.Name(self.env[bind_sym.name], py.Store())
            comp_extra = [0] if PY36 else []  # comprehension:is_async
            comp = py.comprehension(
                store_bind_sym, py.Tuple([self.visit(bind_expr)], py.Load()),
                [], *comp_extra
            )
            expr = py.IfExp(test, self.visit(then_), self.visit(else_))
        gen = py.GeneratorExp(expr, [comp])
        return py.Call(py.Name('next', py.Load()), [gen], [], None, None)

    def visit_each_expr(self, node):
        sym, var, col, body = node.values
        assert isinstance(var, Symbol)
        col_expr = self.visit(col)
        with self.env.push([var.name]):
            var_name = self.env[var.name]
            body_expr = self.visit(body)
        py_comp_extra = [0] if PY36 else []  # comprehension:is_async
        return py.ListComp(body_expr,
                           [py.comprehension(py.Name(var_name, py.Store()),
                                             col_expr, [], *py_comp_extra)])

    def visit_tuple(self, node):
        sym = node.values[0]
        if sym.name == 'get':
            return self.visit_get_expr(node)
        elif sym.name == 'if':
            return self.visit_if_expr(node)
        elif sym.name == 'if_some':
            return self.visit_if_some_expr(node)
        elif sym.name == 'each':
            return self.visit_each_expr(node)
        else:
            values = [self.visit(node) for node in node.values]
            sym_expr, arg_exprs = values[0], values[1:]
            return py.Call(sym_expr, arg_exprs, [], None, None)

    def visit_symbol(self, node):
        if node.name in self.env:
            return self._var_load(self.env[node.name])
        elif isinstance(node.__ref__.to, CallableMeta):
            return self._env_load(node.name)
        else:
            return self._ctx_load(node.name)

    def visit_list(self, node):
        return py.List([self.visit(value) for value in node.values],
                       py.Load())

    def visit_dict(self, node):
        assert not len(node.values) % 2, 'Probably missing keyword value'
        keys = node.values[::2]
        values = node.values[1::2]
        assert all(isinstance(k, Keyword) for k in keys), 'Wrong arguments'
        return py.Dict([py.Str(key.name) for key in keys],
                       [self.visit(value) for value in values])



from __future__ import absolute_import

import re

from itertools import chain
from functools import partial
from collections import namedtuple, OrderedDict

from graphql import print_ast, ast_from_value

from ..graph import Graph, Root, Node, Link, Option, Field, Nothing
from ..graph import GraphTransformer
from ..types import TypeRef, String, Sequence, Boolean, Optional, TypeVisitor
from ..compat import async_wrapper


LIST = namedtuple('LIST', 'of_type')
NON_NULL = namedtuple('NON_NULL', 'of_type')

FieldIdent = namedtuple('FieldIdent', 'node, name')
ArgumentIdent = namedtuple('ArgumentIdent', 'node, field, name')

NAME_RE = re.compile('^[a-zA-Z]\w*$')

ROOT_NAME = 'Root'


class TypeIdent(TypeVisitor):

    def visit_sequence(self, obj):
        return NON_NULL(LIST(self.visit(obj.__item_type__)))

    def visit_optional(self, obj):
        return self.visit(obj.__type__).of_type

    def visit_typeref(self, obj):
        return NON_NULL(obj.__type_name__)

    def visit_string(self, obj):
        return NON_NULL('String')

    def visit_integer(self, obj):
        return NON_NULL('Int')

    def visit_float(self, obj):
        return NON_NULL('Float')

    def visit_boolean(self, obj):
        return NON_NULL('Boolean')


class UnsupportedGraphQLType(TypeError):
    pass


class TypeValidator(TypeVisitor):

    @classmethod
    def is_valid(cls, type_):
        try:
            cls().visit(type_)
        except UnsupportedGraphQLType:
            return False
        else:
            return True

    def visit_any(self, obj):
        raise UnsupportedGraphQLType()


def not_implemented(*args, **kwargs):
    raise NotImplementedError(args, kwargs)


def na_maybe(graph):
    return Nothing


def na_many(graph, ids=None, options=None):
    if ids is None:
        return []
    else:
        return [[] for _ in ids]


def _nodes_map(graph):
    return OrderedDict(chain(((n.name, n) for n in graph.nodes),
                             ((ROOT_NAME, graph.root),)))


def type_link(graph, options):
    name = options['name']
    if name in _nodes_map(graph):
        return name
    else:
        return Nothing


def types_link(graph):
    return [n for n in _nodes_map(graph) if NAME_RE.match(n)]


def type_info(graph, fields, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        if ident in nodes_map:
            node = nodes_map[ident]
            info = {'id': ident,
                    'kind': 'OBJECT',
                    'name': ident,
                    'description': node.description}
        elif isinstance(ident, NON_NULL):
            info = {'id': ident,
                    'kind': 'NON_NULL'}
        elif isinstance(ident, LIST):
            info = {'id': ident,
                    'kind': 'LIST'}
        else:
            assert ident in {'String', 'Int', 'Boolean', 'Float'}, ident
            info = {'id': ident,
                    'name': ident,
                    'kind': 'SCALAR'}
        yield [info.get(f.name) for f in fields]


def validate_field(field):
    if not NAME_RE.match(field.name):
        return False
    if field.type is None or not TypeValidator.is_valid(field.type):
        return False
    for option in field.options:
        if not NAME_RE.match(option.name):
            return False
        if option.type is None or not TypeValidator.is_valid(option.type):
            return False
    return True


def type_fields_link(graph, ids, options):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident]
        field_idents = [FieldIdent(ident, f.name) for f in node.fields
                        if validate_field(f)]
        if not field_idents:
            raise TypeError('Node "{}" does not contain any typed field, which '
                            'is not acceptable for GraphQL in order to define '
                            'schema type'.format(ident))
        yield field_idents


def type_of_type_link(graph, ids):
    for ident in ids:
        if isinstance(ident, (NON_NULL, LIST)):
            yield ident.of_type
        else:
            yield Nothing


def field_info(graph, fields, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.name]
        info = {'id': ident,
                'name': field.name,
                'description': field.description,
                'isDeprecated': False,
                'deprecationReason': None}
        yield [info[f.name] for f in fields]


def field_type_link(graph, ids):
    nodes_map = _nodes_map(graph)
    type_ident = TypeIdent()
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.name]
        yield type_ident.visit(field.type)


def field_args_link(graph, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.name]
        yield [ArgumentIdent(ident.node, field.name, option.name)
               for option in field.options]


def input_value_info(graph, fields, ids):
    nodes_map = _nodes_map(graph)
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.field]
        option = field.options_map[ident.name]
        if option.default is Nothing:
            default = None
        elif option.default is None:
            # graphql-core currently can't parse/print "null" values
            default = 'null'
        else:
            default = print_ast(ast_from_value(option.default))
        info = {'id': ident,
                'name': option.name,
                'description': option.description,
                'defaultValue': default}
        yield [info[f.name] for f in fields]


def input_value_type_link(graph, ids):
    nodes_map = _nodes_map(graph)
    type_ident = TypeIdent()
    for ident in ids:
        node = nodes_map[ident.node]
        field = node.fields_map[ident.field]
        option = field.options_map[ident.name]
        yield type_ident.visit(option.type)


GRAPH = Graph([
    Node('__Type', [
        Field('id', None, type_info),
        Field('kind', String, type_info),
        Field('name', String, type_info),
        Field('description', String, type_info),

        # OBJECT and INTERFACE only
        Link('fields', Sequence[TypeRef['__Field']], type_fields_link,
             requires='name',
             options=[Option('includeDeprecated', Boolean, default=False)]),

        # OBJECT only
        Link('interfaces', Sequence[TypeRef['__Type']], na_many,
             requires='id'),

        # INTERFACE and UNION only
        Link('possibleTypes', Sequence[TypeRef['__Type']], na_many,
             requires='id'),

        # ENUM only
        Link('enumValues', Sequence[TypeRef['__EnumValue']], na_many,
             requires='id',
             options=[Option('includeDeprecated', Boolean, default=False)]),

        # INPUT_OBJECT only
        Link('inputFields', Sequence[TypeRef['__InputValue']], na_many,
             requires='id'),

        # NON_NULL and LIST only
        Link('ofType', Optional[TypeRef['__Type']], type_of_type_link,
             requires='id'),
    ]),
    Node('__Field', [
        Field('id', None, field_info),
        Field('name', String, field_info),
        Field('description', String, field_info),

        Link('args', Sequence[TypeRef['__InputValue']], field_args_link,
             requires='id'),
        Link('type', TypeRef['__Type'], field_type_link, requires='id'),
        Field('isDeprecated', Boolean, field_info),
        Field('deprecationReason', String, field_info),
    ]),
    Node('__InputValue', [
        Field('id', None, input_value_info),
        Field('name', String, input_value_info),
        Field('description', String, input_value_info),
        Link('type', TypeRef['__Type'], input_value_type_link, requires='id'),
        Field('defaultValue', String, input_value_info),
    ]),
    Node('__Directive', [
        Field('name', String, not_implemented),
        Field('description', String, not_implemented),
        Field('locations', String, not_implemented),  # FIXME: Sequence[String]
        Link('args', Sequence[TypeRef['__InputValue']], not_implemented,
             requires=None),
    ]),
    Node('__EnumValue', [
        Field('name', String, not_implemented),
        Field('description', String, not_implemented),
        Field('isDeprecated', Boolean, not_implemented),
        Field('deprecationReason', String, not_implemented),
    ]),
    Root([
        Node('__schema', [
            Link('types', Sequence[TypeRef['__Type']], types_link,
                 requires=None),
            Link('queryType', TypeRef['__Type'], lambda graph: ROOT_NAME,
                 requires=None),
            Link('directives', Sequence[TypeRef['__Directive']], na_many,
                 requires=None),
            Link('mutationType', Optional[TypeRef['__Type']], na_maybe,
                 requires=None),
            Link('subscriptionType', Optional[TypeRef['__Type']], na_maybe,
                 requires=None),
        ]),
        Link('__type', Optional[TypeRef['__Type']], type_link, requires=None,
             options=[Option('name', String)]),
    ]),
])


class BindToGraph(GraphTransformer):

    def __init__(self, graph):
        self.graph = graph

    def visit_field(self, obj):
        field = super(BindToGraph, self).visit_field(obj)
        field.func = partial(field.func, self.graph)
        return field

    def visit_link(self, obj):
        link = super(BindToGraph, self).visit_link(obj)
        link.func = partial(link.func, self.graph)
        return link


class MakeAsync(GraphTransformer):

    def visit_field(self, obj):
        field = super(MakeAsync, self).visit_field(obj)
        field.func = async_wrapper(field.func)
        return field

    def visit_link(self, obj):
        link = super(MakeAsync, self).visit_link(obj)
        link.func = async_wrapper(link.func)
        return link


def type_name_field_func(node_name, fields, ids=None):
    return [[node_name] for _ in ids] if ids is not None else [node_name]


class AddIntrospection(GraphTransformer):

    def __init__(self, introspection_graph, type_name_field_factory):
        self.introspection_graph = introspection_graph
        self.type_name_field_factory = type_name_field_factory

    def visit_node(self, obj):
        node = super(AddIntrospection, self).visit_node(obj)
        node.fields.append(self.type_name_field_factory(obj.name))
        return node

    def visit_root(self, obj):
        root = super(AddIntrospection, self).visit_root(obj)
        root.fields.append(self.type_name_field_factory(ROOT_NAME))
        return root

    def visit_graph(self, obj):
        graph = super(AddIntrospection, self).visit_graph(obj)
        graph.items.extend(self.introspection_graph.items)
        return graph


class GraphQLIntrospection(GraphTransformer):

    def __type_name__(self, node_name):
        return Field('__typename', String,
                     partial(type_name_field_func, node_name))

    def __introspection_graph__(self, graph):
        return BindToGraph(graph).visit(GRAPH)

    def visit_node(self, obj):
        node = super(GraphQLIntrospection, self).visit_node(obj)
        node.fields.append(self.__type_name__(obj.name))
        return node

    def visit_root(self, obj):
        root = super(GraphQLIntrospection, self).visit_root(obj)
        root.fields.append(self.__type_name__(ROOT_NAME))
        return root

    def visit_graph(self, obj):
        introspection_graph = self.__introspection_graph__(obj)
        items = [self.visit(node) for node in obj.items]
        items.extend(introspection_graph.items)
        return Graph(items)


class AsyncGraphQLIntrospection(GraphQLIntrospection):

    def __type_name__(self, node_name):
        return Field('__typename', String,
                     async_wrapper(partial(type_name_field_func, node_name)))

    def __introspection_graph__(self, graph):
        introspection_graph = BindToGraph(graph).visit(GRAPH)
        introspection_graph = MakeAsync().visit(introspection_graph)
        return introspection_graph
